use std::sync::Arc;

use eyre::Result;
use foxchat::{
    s2s::{Dispatch, Payload},
    signature,
};
use futures::{
    stream::{SplitSink, SplitStream},
    SinkExt, StreamExt,
};
use reqwest::Url;
use tokio::{
    net::TcpStream,
    sync::{broadcast, mpsc, RwLock},
};
use tokio_tungstenite::{connect_async, tungstenite::Message, MaybeTlsStream, WebSocketStream};

use crate::{app_state::AppState, model::chat_instance::ChatInstance};

pub enum ServerSocketMessage {
    Connect,
    Disconnect,
}

pub struct SocketState {
    id: String,
    domain: String,
    conn_count: u32,
    broadcast: broadcast::Sender<Dispatch>,
}

impl SocketState {
    pub async fn new(
        state: Arc<AppState>,
        instance: &ChatInstance,
    ) -> Result<mpsc::Sender<ServerSocketMessage>> {
        let (msg_tx, msg_rx) = mpsc::channel::<ServerSocketMessage>(10);

        let socket_state = Arc::new(RwLock::new(SocketState {
            id: instance.id.clone(),
            domain: instance.domain.clone(),
            conn_count: 0,
            broadcast: state.broadcast.clone(),
        }));

        let url = Url::parse(&format!("{}/_fox/chat/ws", instance.base_url))?;
        let (stream, _) = connect_async(url).await?;

        let (tx, rx) = stream.split();

        tokio::spawn(read(socket_state.clone(), rx, msg_tx.clone()));
        tokio::spawn(write(state.clone(), socket_state.clone(), tx, msg_rx));

        Ok(msg_tx)
    }
}

async fn write(
    app_state: Arc<AppState>,
    state: Arc<RwLock<SocketState>>,
    mut tx: SplitSink<WebSocketStream<MaybeTlsStream<TcpStream>>, Message>,
    msg_rx: mpsc::Receiver<ServerSocketMessage>,
) {
    // Create signature and turn it into JSON
    let host = state.read().await.domain.clone();
    let (sig, date) =
        signature::build_signature(&app_state.private_key, &host, "/_fox/chat/ws", None, None);
    let msg = match to_json(Payload::Identify {
        host: state.read().await.domain.clone(),
        date: signature::format_date(date),
        server: app_state.config.domain.clone(),
        signature: sig,
    }) {
        Ok(t) => t,
        Err(e) => {
            tracing::error!("Error serializing IDENTIFY payload: {}", e);
            tx.close().await.ok();
            return;
        }
    };

    // Send IDENTIFY message
    match tx.send(msg).await {
        Ok(_) => {}
        Err(e) => {
            tracing::error!("Error sending IDENTIFY payload: {}", e);
            tx.close().await.ok();
            return;
        }
    }
}

async fn read(
    state: Arc<RwLock<SocketState>>,
    rx: SplitStream<WebSocketStream<MaybeTlsStream<TcpStream>>>,
    msg_tx: mpsc::Sender<ServerSocketMessage>,
) {
}

fn to_json(p: Payload) -> Result<Message> {
    Ok(Message::Text(serde_json::to_string(&p)?))
}
